// --------------------------------------------------------------------
//  DDE.C
// --------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#define INCL_DOSSEMAPHORES
#define INCL_DOSFILEMGR
#define INCL_WINMESSAGEMGR
#define INCL_WINWINDOWMGR
#define INCL_WINDIALOGS
#define INCL_WINLISTBOXES
#define INCL_WINBUTTONS
#define INCL_WINDDE

#include <os2.h>

#include "ddedlg.h"

HAB hab;
HWND dde_hwndDlg= 0;
HWND dde_hwndZocDDE= 0;

HEV _dde_hevWaitForDdeInit;
HEV _dde_hevWaitForAckOrData;
int _dde_current_ack;
char *_dde_current_data;



void _Optlink work_Query(void*);
MRESULT EXPENTRY _dde_DlgProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
DDESTRUCT *MakeObject(HWND, char *, USHORT, VOID *, USHORT);
int ZocCommand(HWND hwndS, HWND hwndC, char* cmd);
char *ZocFunction(HWND hwndS, HWND hwndC, char* req);
void LineOut(HWND, char *txt);
void ErrorMsg(char *txt);


// --------------------------------------------------------------------
//  MAIN FUNC
// --------------------------------------------------------------------
int main(void)
{
	HMQ  hmq;
	QMSG qmsg;

	hab= WinInitialize(0);
	hmq= WinCreateMsgQueue(hab, 64);

	DosCreateEventSem(NULL, &_dde_hevWaitForDdeInit, 0, FALSE);
	DosCreateEventSem(NULL, &_dde_hevWaitForAckOrData, 0, FALSE);
	
	//
	// create main window (loaded from ressource)
	//
	dde_hwndDlg= WinLoadDlg(HWND_DESKTOP, 0L, 
							_dde_DlgProc, 0L, 
							DLG_WIN_MAIN, NULL); 
	//
	// std msg loop 
	//
	if (dde_hwndDlg)
	{
		while(WinGetMsg(hab, &qmsg, 0L, 0, 0))
			WinDispatchMsg(hab, &qmsg);
		
		WinDestroyWindow(dde_hwndDlg);
	}

	//
	// tidy up
	//
	WinDestroyMsgQueue(hmq);
	WinTerminate(hab);

	return(0);
}


// --------------------------------------------------------------------
//  WINDOW FUNCTION TO MANAGE THE DIALOG WINDOW 
// --------------------------------------------------------------------
MRESULT EXPENTRY 
_dde_DlgProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
	DDESTRUCT *ddestruct= (DDESTRUCT*)mp2;

	switch(msg)
	{

		/////////////////////////////////////////// 
		// WHEN STARTING WE CONNECT TO ZOC       //
		///////////////////////////////////////////

		case WM_INITDLG:
		{
			CONVCONTEXT cc= {0, };
			
			// set context values
			cc.cb= sizeof(cc);
			cc.idCountry= 49;
			cc.usCodepage= 850;
		
			WinDdeInitiate(hwnd, "ZOC", "COMM-DEBUG", &cc);

			LineOut(hwnd, "** SEND DDE-INIT");

			break;
		}


		/////////////////////////////////////////// 
		// THE RUN BUTTON STARTS THE WORK THREAD //
		///////////////////////////////////////////

		case WM_COMMAND:
		{
			int command = SHORT1FROMMP(mp1);

			switch (command) {
				case DLG_PUSH_RUN: 
						_beginthread(work_Query, NULL, 16384, NULL);	
						break;
			}
			break;
		}


		/////////////////////////////////////////// 
		// WINDOW CLOSE TERMINATES DDE           //
		///////////////////////////////////////////

		case WM_CLOSE:
		{
			WinDdePostMsg(dde_hwndZocDDE, hwnd, WM_DDE_TERMINATE, NULL, 1);
			WinPostMsg(hwnd, WM_QUIT, NULL, NULL);
			goto DefProc;
		}



		/////////////////////////////////////////// 
		//                                       //
		// DDE MESSAGE HANDLING                  //
		//                                       //
		///////////////////////////////////////////

		case WM_DDE_INITIATEACK:
		{
			// mp1 is ZOC's window
			dde_hwndZocDDE= (HWND)mp1;

			// mark Init-Done
			DosPostEventSem(_dde_hevWaitForDdeInit);

			// free dde-msg
			DosFreeMem((void*)mp2);

			LineOut(hwnd, "** GET DDE-INITIATE-ACK");
			break;
		}

		case WM_DDE_ACK:
		{
			// zoc sends the command-rc in the upper 8 bytes of fsStatus
			_dde_current_ack= ddestruct->fsStatus >> 8;

			// ok, ack returned (for ZocCommand() or ZocFunction() below)
			DosPostEventSem(_dde_hevWaitForAckOrData);

			// free dde-msg
			DosFreeMem((void*)mp2);
			LineOut(hwnd, "** GET DDE-ACK");
			break;
		}

		case WM_DDE_DATA:
		{
			LineOut(hwnd, "** GET DDE-DATA");

			// make a copy of the memory buffer for our own use
			_dde_current_data= strdup((char*)mp2+ddestruct->offabData);
			
			// ACK not neccesary, cause ZOC doesn't care about it

			// ok, data returned (for ZocCommand() or ZocFunction() below)
			DosPostEventSem(_dde_hevWaitForAckOrData);

			// free dde-msg
			DosFreeMem((void*)mp2);
			break;
		}

		case WM_DDE_TERMINATE:
		{
			LineOut(hwnd, "** GET DDE-TERMINATE");
			WinPostMsg(hwnd, WM_CLOSE, NULL, NULL);
			break;
		}


		//
		// ALL OTHER STUFF
		//

		DefProc:
		default:
			return WinDefDlgProc(hwnd, msg, mp1, mp2);
	}
	
	return (MRESULT)FALSE;
}


// --------------------------------------------------------------------
//  process that little bugger 
// --------------------------------------------------------------------
void _Optlink work_Query(void* v)
{
	int rc;
	char *result, 
			resultbuf[80]= "",
			cmd[80];

	// init pm handling
	HAB hab2= WinInitialize(0);
	HMQ hmq2= WinCreateMsgQueue(hab2, 64);

	//
	// get values from dialog (3 keywords, and/or selection)
	//
	WinQueryWindowText(
					WinWindowFromID(dde_hwndDlg, DLG_ENTRY_CMD),
					sizeof(cmd), cmd);


	if (strchr(cmd, '(')==NULL) {		// it is a COMMAND
		rc= ZocCommand(dde_hwndZocDDE, dde_hwndDlg, cmd);
		sprintf(resultbuf, "AppRC= %d", rc);
	}
	else {
		result= ZocFunction(dde_hwndZocDDE, dde_hwndDlg, cmd);
		strncpy(resultbuf, result, sizeof(resultbuf)-1);
	}

	// show the result
	WinSetWindowText(WinWindowFromID(dde_hwndDlg, DLG_ENTRY_RESULT), resultbuf);

	WinSetFocus(HWND_DESKTOP, WinWindowFromID(dde_hwndDlg, DLG_ENTRY_CMD));

	// tidy up
	WinDestroyMsgQueue(hmq2);
	WinTerminate(hab2);
}


// --------------------------------------------------------------------
//  Send a command to ZOC and return ZOC's command-rc
// --------------------------------------------------------------------
int ZocCommand(HWND hwndS, HWND hwndC, char* cmd)
{
	int rc;
	ULONG dummy;
	DDESTRUCT *pDde;

	// don't do anything before INIT was done
	rc= DosWaitEventSem(_dde_hevWaitForDdeInit, 2000);
	if (rc!=0) {
		ErrorMsg("DDE was not initialized!");
		return -1;
	}

	// compose a message with the command-request
	pDde= MakeObject(hwndS, "", DDE_FACKREQ, cmd, strlen(cmd));

	// send message and wait for ACK
	DosResetEventSem(_dde_hevWaitForAckOrData, &dummy);
	WinDdePostMsg(hwndS, hwndC, WM_DDE_EXECUTE, pDde, 1);	
	LineOut(dde_hwndDlg, "** SEND DDE-EXECUTE");
	DosWaitEventSem(_dde_hevWaitForAckOrData, SEM_INDEFINITE_WAIT);

	// return ACK-rc
	return (int)(signed char)_dde_current_ack;
}


// --------------------------------------------------------------------
//  Send a request to ZOC and return a pointer to the result
//  (pointer only valid until the function is called again)
// --------------------------------------------------------------------
char *ZocFunction(HWND hwndS, HWND hwndC, char* req)
{
	int rc;
	ULONG dummy;
	DDESTRUCT *pDde;

	// don't do anything before INIT was done
	rc= DosWaitEventSem(_dde_hevWaitForDdeInit, 2000);
	if (rc!=0) {
		ErrorMsg("DDE was not initialized!");
		return "DDE-ERROR";
	}

	// compose a message with the command-request
	pDde= MakeObject(hwndS, req, 0, NULL, 0);	

	// probably have to delete the old pointer
	if (_dde_current_data) {
		free(_dde_current_data);
		_dde_current_data= NULL;
	}

	// send the request and wait for the result DATA-packet
	DosResetEventSem(_dde_hevWaitForAckOrData, &dummy);
	WinDdePostMsg(hwndS, hwndC, WM_DDE_REQUEST, pDde, 1);	
	LineOut(dde_hwndDlg, "** SEND DDE-REQUEST");
	DosWaitEventSem(_dde_hevWaitForAckOrData, SEM_INDEFINITE_WAIT);

	// return the result pointer
	return _dde_current_data;
}


// --------------------------------------------------------------------
//  compose a DDE packet (see IBM OS/2 Progr. Ref. Vol. II)
// --------------------------------------------------------------------
DDESTRUCT *
MakeObject(HWND hwnd, char *itemname, USHORT fsStatus, 
				VOID *pdata, USHORT datalen)
{
	DDESTRUCT *pDde= NULL;
	ULONG rc, itemlen;
	VOID *sharedmem= NULL;
	PID pid;
	TID tid;

	if (itemname != NULL)
		itemlen= strlen(itemname)+1;
	else
		itemlen= 1;

	rc= DosAllocSharedMem(&sharedmem, NULL,
							sizeof(DDESTRUCT)+itemlen+datalen,
							PAG_COMMIT | PAG_READ | PAG_WRITE | OBJ_GIVEABLE);
	if (!rc) {

		memset(sharedmem, 0, sizeof(DDESTRUCT)+itemlen+datalen);

		WinQueryWindowProcess(hwnd, &pid, &tid);
		DosGiveSharedMem(&sharedmem, pid, PAG_READ|PAG_WRITE);

		// Initialize DDESTRUCT
		pDde= (PDDESTRUCT) sharedmem;
		pDde->cbData= (LONG) datalen;
		pDde->fsStatus= fsStatus;
		pDde->usFormat= DDEFMT_TEXT;
		pDde->offszItemName= sizeof(DDESTRUCT);

		if (datalen && pdata!=NULL)
			pDde->offabData= sizeof(DDESTRUCT)+itemlen;
		else
			pDde->offabData= 0;
															
		// Copy item name immediately following DDESTRUCT
		if (itemname!=NULL)
			strcpy(DDES_PSZITEMNAME(pDde), itemname);
		else
			strcpy(DDES_PSZITEMNAME(pDde), "");
															
		// Copy data immediately following item name
		if (pdata != NULL)
			memcpy(DDES_PABDATA(pDde), pdata, datalen);
															
	}

	return (pDde);
}



// --------------------------------------------------------------------
//  write a line of text to the listbox with id  DLG_LIST_MSGS
// --------------------------------------------------------------------
void
LineOut(HWND hwnd, char *txt)
{
	int rc;

	if (txt) {
		rc= (int)WinSendDlgItemMsg(hwnd, DLG_LIST_MSGS, 
									LM_INSERTITEM, 
									MPFROMSHORT(LIT_END), 
									txt);
		if (rc>5)
			WinSendDlgItemMsg(hwnd, DLG_LIST_MSGS, 
									LM_SETTOPINDEX,
									MPFROMLONG(rc-4), 0L);
	}
	else 
		WinSendMsg(WinWindowFromID(hwnd, DLG_LIST_MSGS), 
									LM_DELETEALL, 0L, 0L);
}


// --------------------------------------------------------------------
//  a simple error message window
// --------------------------------------------------------------------
void 
ErrorMsg(char *txt)
{
	extern HWND dde_hwndDlg;

	WinMessageBox(HWND_DESKTOP, 
					dde_hwndDlg, txt, "DDE Sample", 0, 
					MB_OK | MB_ERROR | MB_MOVEABLE);
}

